﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace Hilbert
{
    enum DataType
    {
        TRAIN, SELECT, TEST
    }

    class Program
    {
        private static List<double[]> _inputs = new List<double[]>();
        private static List<double> _target = new List<double>();
        private static List<DataType> _dt = new List<DataType>();
        private static List<U> _ulist = new List<U>();
        private static U _bigU = new U();
        private static Random _rnd = new Random();
        private static double _targetMin;
        private static double _targetMax;

        static private void Swap(ref int n1, ref int n2)
        {
            int v1 = n1;
            n1 = n2;
            n2 = v1;
        }

        public static int[] Shuffle(int size)
        {
            int[] position = new int[size];
            for (int i = 0; i < size; ++i)
            {
                position[i] = i;
            }
            for (int i = 0; i < position.Length; ++i)
            {
                int plusPos = _rnd.Next() % position.Length;
                int next = i + plusPos;
                if (next > position.Length - 1) next -= (position.Length - 1);

                int el1 = position[i];
                int el2 = position[next];
                Swap(ref el1, ref el2);
                position[i] = el1;
                position[next] = el2;
            }
            return position;
        }

        static double[] Convert(double[] x)
        {
            int Levels = 20;
            int[] positions = Shuffle(Levels);
            double[] y = new double[x.Length];
            double xmin = x.Min();
            double xmax = x.Max();
            double delta = (xmax - xmin) / (Levels - 1);
            for (int i = 0; i < x.Length; ++i)
            {
                int level = (int)Math.Round((x[i] - xmin) / delta);
                y[i] = positions[level];
            }
            double ymin = y.Min();
            double ymax = y.Max();
            for (int i = 0; i < y.Length; ++i)
            {
                y[i] = (y[i] - ymin) / (ymax - ymin) * (xmax - xmin) + xmin;
            }
            return y;
        }

        static void GenerateInitialOperators(int M, int nx)
        {
            _ulist.Clear();
            //find min, max
            int size = _inputs[0].Length;
            double[] min = new double[size];
            double[] max = new double[size];
            for (int k = 0; k < size; ++k)
            {
                min[k] = _inputs[0][k];
                max[k] = _inputs[0][k];
            }
            foreach (double[] data in _inputs)
            {
                for (int k = 0; k < size; ++k)
                {
                    if (data[k] < min[k]) min[k] = data[k];
                    if (data[k] > max[k]) max[k] = data[k];
                }
            }
            //for (int m = 0; m < size; ++m)
            //{
            //    Console.WriteLine(min[m] + " " + max[m]);
            //}
            /////// end

            for (int counter = 0; counter < M; ++counter)
            {
                double[] x = new double[_target.Count];
                for (int i = 0; i < _target.Count; ++i)
                {
                    x[i] = _target[i];
                }
                double[] z = Convert(x);

                U u = new U();
                u.Initialize(_inputs[0].Length, nx, min, max);
                for (int count = 1; count < 10; ++count)
                {
                    double accuracy = 0.0;
                    int cnt = 0;
                    for (int i = 0; i < _inputs.Count; ++i)
                    {
                        if (DataType.TRAIN != _dt[i]) continue;
                        double[] data = _inputs[i];
                        double diff = z[i] - u.GetU(data);
                        u.Update(diff, data, 0.05);
                        accuracy += diff * diff;
                        ++cnt;
                    }                    
                    accuracy /= cnt;
                    accuracy = Math.Sqrt(accuracy);
                    accuracy /= (_targetMax - _targetMin);
                    //Console.WriteLine("Error for initial operator {0}, {1:0.0000}", counter, accuracy); 
                }
                _ulist.Add(u);
            }
        }

        static double[] GetVector(double[] data)
        {
            int size = _ulist.Count;
            double[] vector = new double[size];
            for (int i = 0; i < size; ++i)
            {
                vector[i] = _ulist[i].GetU(data);
            }
            return vector;
        }

        static void BuildRepresentation(int M, int nx)
        {
            double[] min = new double[M];
            double[] max = new double[M];
            for (int i = 0; i < M; ++i)
            {
                min[i] = _targetMin;
                max[i] = _targetMax;
            }
            _bigU.Clear();
            _bigU.Initialize(M, nx, min, max);
            int N = _inputs.Count;
            double error = (_targetMax - _targetMin) / 150.0;
            double training_error = 0.0;

            for (int step = 0; step < 1000; ++step)
            {
                double RMSE = 0.0;
                int cnt = 0;
                for (int i = 0; i < N; ++i)
                {
                    if (DataType.TRAIN != _dt[i]) continue;
                    double[] data = _inputs[i];
                    double[] v = GetVector(data);
                    double diff = _target[i] - _bigU.GetU(v);

                    if (Math.Abs(diff) > error)
                    {
                        for (int k = 0; k < _ulist.Count; ++k)
                        {
                            double derrivative = _bigU.GetDerrivative(k, v[k]);
                            if (Math.Abs(derrivative) < 0.1)
                            {
                                if (derrivative < 0.0) derrivative = -0.1;
                                else derrivative = 0.1;
                            }
                            _ulist[k].Update(diff / derrivative, data, 0.05);
                        }
                        v = GetVector(data);
                        diff = _target[i] - _bigU.GetU(v);
                        _bigU.Update(diff, v, 0.05);
                    }
                    RMSE += diff * diff;
                    ++cnt;
                }

                RMSE /= cnt;
                RMSE = Math.Sqrt(RMSE);
                RMSE /= (_targetMax - _targetMin);
                //Console.WriteLine("Error in big operator {0:0.0000}", accuracy);
                training_error = RMSE;
            }
            //Console.WriteLine("Training is completed, relative error {0:0.0000}", training_error);
        }

        static void ReadDataAirfoilSelfNoise()
        {
            string fileName = @"..\..\airfoil_self_noise.csv";
            if (!File.Exists(fileName))
            {
                Console.WriteLine("File not found {0}", fileName);
                Environment.Exit(0);
            }
            string[] lines = System.IO.File.ReadAllLines(fileName);
            for (int m = 1; m < lines.Length; ++m) //we skip first line
            {
                string[] sdata = lines[m].Split(';');
                if (6 != sdata.Length)
                {
                    Console.WriteLine("Misformatted data");
                    Environment.Exit(0);
                }
                double[] data = new double[5];
                for (int i = 0; i < data.Length; ++i)
                {
                    Double.TryParse(sdata[i], out data[i]);
                }
                double target = 0.0;
                Double.TryParse(sdata[5], out target);
                _inputs.Add(data);
                _target.Add(target);

                if (_rnd.Next() % 100 <= 60) _dt.Add(DataType.TRAIN);
                else
                {
                    if (_rnd.Next() % 100 <= 50) _dt.Add(DataType.SELECT);
                    else _dt.Add(DataType.TEST);
                }
            }

            _targetMin = _target.Min();
            _targetMax = _target.Max();
  
            //Stat of data sample
            int nTrain = 0;
            int nSelect = 0;
            int nTest = 0;
            int nTotal = 0;
            foreach (DataType dt in _dt)
            {
                if (DataType.TRAIN == dt) ++nTrain;
                else if (DataType.SELECT == dt) ++nSelect;
                else if (DataType.TEST == dt) ++nTest;
                ++nTotal;
            }
            Console.WriteLine("Data is ready.");
            Console.WriteLine("Total = {0}, training sample = {1}, selection sample = {2}, test sample = {3}", 
                nTotal, nTrain, nSelect, nTest);
        }

        static double DoSelection()
        {
            double RMSE = 0.0;
            int cnt = 0;
            int N = _inputs.Count;
            for (int i = 0; i < N; ++i)
            {
                if (DataType.SELECT != _dt[i]) continue;
                double[] data = _inputs[i];
                double[] v = GetVector(data);
                double diff = _target[i] - _bigU.GetU(v);
                RMSE += diff * diff;
                ++cnt;
            }
            RMSE /= cnt;
            RMSE = Math.Sqrt(RMSE);
            RMSE /= (_targetMax - _targetMin);
            return RMSE;
        }

        static double DoTest()
        {
            double RMSE = 0.0;
            int cnt = 0;
            int N = _inputs.Count;
            for (int i = 0; i < N; ++i)
            {
                if (DataType.TEST != _dt[i]) continue;
                double[] data = _inputs[i];
                double[] v = GetVector(data);
                double diff = _target[i] - _bigU.GetU(v);
                RMSE += diff * diff;
                ++cnt;
            }
            RMSE /= cnt;
            RMSE = Math.Sqrt(RMSE);
            RMSE /= (_targetMax - _targetMin);
            return RMSE;
        }

        static double ComputeCorrelationCoeff()
        {
            int N = _inputs.Count;
            double[] targetEstimate = new double[N];
            int count = 0;
            for (int i = 0; i < N; ++i)
            {
                double[] data = _inputs[i];
                double[] v = GetVector(data);
                targetEstimate[i] = _bigU.GetU(v);
                if (DataType.TEST == _dt[i]) ++count;
            }
            double[] x = new double[count];
            double[] y = new double[count];
            count = 0;
            for (int i = 0; i < N; ++i)
            {
                if (DataType.TEST == _dt[i])
                {
                    x[count] = targetEstimate[i];
                    y[count] = _target[i];
                    ++count;
                }
            }
            return PearsonCorrelation(x, y);
       }

        static double PearsonCorrelation(double[] x, double[] y)
        {
            double xy = 0.0;
            double x2 = 0.0;
            double y2 = 0.0;
            for (int i = 0; i < x.Length; ++i)
            {
                xy += x[i] * y[i];
                x2 += x[i] * x[i];
                y2 += y[i] * y[i];
            }
            xy /= (double)(x.Length);
            x2 /= (double)(x.Length);
            y2 /= (double)(y.Length);
            double xav = x.Average();
            double yav = y.Average();
            double ro = xy - xav * yav;
            ro /= Math.Sqrt(x2 - xav * xav);
            ro /= Math.Sqrt(y2 - yav * yav);
            return ro;
        }

        static void Main(string[] args)
        {
            ReadDataAirfoilSelfNoise();
            int M = 11;
            int nx = 10;
            for (int count = 0; count < 10; ++count)
            {
                GenerateInitialOperators(M, nx);
                BuildRepresentation(M, nx);
                double selectionError = DoSelection();
                double pearson = ComputeCorrelationCoeff();
                Console.WriteLine("Run = {0}, selection sample error = {1:0.0000}, test sample pearson correlation coeff. = {2:0.0000}",
                    count, selectionError, pearson);
            }
        }
    }
}
